/**
  * Java Service Launcher
  *
*/

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <time.h>

#include "service.h"

//service name
LPTSTR szappname;
LPTSTR szservicename;
LPTSTR szdisplayname;
LPTSTR szconsolehandler;
LPTSTR szdependencies;
LPTSTR szLoadOrderGroup;
LPTSTR szAccount;
LPTSTR szPwd;
DWORD dwStartType;

LPTSTR stopClass;
LPTSTR stopMethod;
LPTSTR stopSignature;
LPTSTR path;

int stopport;
int global_argc;
int startdelay = 0;
LPTSTR  global_jrePath;
LPTSTR  global_wrkdir;
LPTSTR* global_argv;

// internal variables
SERVICE_STATUS          ssStatus;       // current status of the service
SERVICE_STATUS_HANDLE   sshStatusHandle;
DWORD                   dwErr = 0;
BOOL                    bDebug = FALSE;
BOOL                    bCommandLine = FALSE;
BOOL					bService = FALSE;
BOOL					bCommandLineDebug = FALSE;
BOOL					useConsoleHandler =TRUE;
BOOL					bPaused = FALSE;
TCHAR                   szErr[256];

//p0.99
//redirected stdout and stderr
FILE* restdout = NULL;
FILE* restderr = NULL;
LPTSTR stdoutfile = NULL;
LPTSTR stderrfile = NULL;
int	exceptionErrorCode = 0; 
LPTSTR onExitError = "";
LPTSTR szModulePath = "";

// internal function prototypes
VOID WINAPI service_ctrl(DWORD dwCtrlCode);
VOID WINAPI service_main(DWORD dwArgc, LPTSTR *lpszArgv);
VOID CmdInstallService();
VOID CmdConfigureService();
VOID CmdRemoveService();
VOID CmdDebugService(int argc, LPTSTR* argv);
VOID CmdRunService(int argc, LPTSTR* argv);
BOOL WINAPI ControlHandler ( DWORD dwCtrlType );
//BOOL WINAPI ControlHandler2 ( DWORD dwCtrlType );
LPTSTR GetLastErrorText( LPTSTR lpszBuf, DWORD dwSize );

//
//  FUNCTION: main
//
//  PURPOSE: entrypoint for service
//
//  PARAMETERS:
//    argc - number of command line arguments
//    argv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    main() either performs the command line task, or
//    call StartServiceCtrlDispatcher to register the
//    main service thread.  When the this call returns,
//    the service has stopped, so exit.
//
void _CRTAPI1 main(int argc, LPTSTR* argv)
{
    SERVICE_TABLE_ENTRY dispatchTable[] =
    {
        { TEXT(SZSERVICENAME), (LPSERVICE_MAIN_FUNCTION)service_main },
        { NULL, NULL }
    };

	if ( bDebug )
		logf( "started" );

    if ( (argc > 1) &&
         ((*argv[1] == '-') || (*argv[1] == '/')) )
    {
        if ( _stricmp( "install", argv[1]+1 ) == 0 )
        {
			getParams( &argc, &argv, FALSE, TRUE );
			dispatchTable[0].lpServiceName = szservicename;
            CmdInstallService();
        }
		else if ( _stricmp( "configure", argv[1]+1 ) == 0 )
        {
			getParams( &argc, &argv, FALSE, TRUE );
			dispatchTable[0].lpServiceName = szservicename;
            CmdConfigureService();
        }
        else if ( _stricmp( "remove", argv[1]+1 ) == 0 )
        {
			getParams( &argc, &argv, FALSE, TRUE );
			dispatchTable[0].lpServiceName = szservicename;
            CmdRemoveService();
        }
        else if ( _stricmp( "debug", argv[1]+1 ) == 0 )
        {
			getParams( &argc, &argv, FALSE, TRUE );
			dispatchTable[0].lpServiceName = szservicename;
            bCommandLine = TRUE;
			bCommandLineDebug = TRUE;
            CmdDebugService(argc, argv);
        }
		else if ( _stricmp( "run", argv[1]+1 ) == 0 )
        {
          //16.05.01 Erwin Vervondel
          //Allow to run as a standalone application
			    getParams( &argc, &argv, FALSE, FALSE );
			    dispatchTable[0].lpServiceName = szservicename;
          bCommandLine = TRUE;
          CmdRunService(argc, argv);
        }
        else
        {
            goto dispatch;
        }
        exit(0);
    }

    // if it doesn't match any of the above parameters
    // the service control manager may be starting the service
    // so we must call StartServiceCtrlDispatcher
    dispatch:
						      
		printf( "JSL Java Service Launcher by Michael Roeschter (Michael@Roeschter.com)\nVersion 0.99b May, 5th 2002\n" );
		printf( "Command line options:\n" );
        printf( "%s -install [ini]    to install the service\n", argv[0] );
		printf( "%s -configure [ini]  to reconfigure an installed service with new dependencies\n", argv[0] );
        printf( "%s -remove [ini]     to remove the service\n", argv[0] );
        printf( "%s -debug [ini]      to run as a console app for debugging\n", argv[0] );
		printf( "%s -run [ini]        to run as a standalone console app\n", argv[0] );
		
		//Read parameters from the ini file for this service	
		getParams( &argc, &argv, TRUE, TRUE);

		global_argc = argc;
		global_argv = argv;

		if ( bDebug )
			logf( "params read, calling dispatcher" );		

        printf( "\nStartServiceCtrlDispatcher being called.\n" );
        printf( "This may take several seconds.  Please wait.\n" );
		
		bService = TRUE;
		dispatchTable[0].lpServiceName = szservicename;	

        if (!StartServiceCtrlDispatcher(dispatchTable))
		{
            AddToMessageLog(TEXT("StartServiceCtrlDispatcher failed."));
			if ( bDebug )
			{
				logf( GetLastErrorText(szErr, 256) );
				logf( "StartServiceCtrlDispatcher failed" );
			}
		}

		if ( bDebug )
			logf( "StartServiceCtrlDispatcher returned" );
}


/**
  * Read parameters from the configuration file 
*/
void getParams(int *pargc, LPTSTR** pargv, boolean show, boolean runAsService)
{
    int argc = *pargc;
    LPTSTR* argv = *pargv;
    int nargc; 
    LPTSTR* nargv;
    int i;

	LPTSTR _java = "java";
	LPTSTR _service = "service";
	int params;
	LPTSTR _param = "param00";

	int bsize = 4096;
	LPTSTR buffer = malloc( bsize );	
	LPTSTR commandLine = malloc( bsize );
	LPTSTR cLine = commandLine;

	int pos = 0;
	int cpos = 0;
	int size = 0;
	int currentParam = 0;
	int startParam = 0;
	char* stmp;

	//Number of static parameters
	int STATIC_PARAMS = 4;

	//p0.99
	TCHAR cszPath[512 * sizeof( TCHAR )];
	LPTSTR szPath = cszPath;

	TCHAR szDrive[_MAX_DRIVE];// New 24.06.1999, Heiner Jostkleigrewe
	TCHAR szDir[_MAX_DIR];    // New 24.06.1999, Heiner Jostkleigrewe	

	//p0.99
	//Specify the location of the ini file
	if ( argc >= 3 )
	{
		szPath = argv[2];
	} else
	{
		getIniFile( szPath );
	}
	getIniDir( szPath, szDrive, szDir ); //New 24.06.1999 Heiner Jostkleigrewe

	
	if( ( fopen( szPath, "r" )) == NULL )
	{
      printf( "Initialization file %s not found\n", szPath );
	  exit(0);
	}

	if ( show )
	  printf( "\nInitalization file: %s\n", szPath );

	//Load the initialization parameters

	//param count
	params = GetPrivateProfileInt( _java, "params", 0, szPath );

	if (runAsService)
	{
	  nargc = params + STATIC_PARAMS;// Changed 24.06.1999, Heiner Jostkleigrewe
	}
	else
	{
	  nargc = params + 1;// Changed 16.05.2001, Erwin Vervondel
	}
	nargv = malloc(( nargc ) * sizeof(LPTSTR));// Changed 24.06.1999, Heiner Jostkleigrewe

	*pargc = nargc;
    *pargv = nargv;
	nargv[currentParam++] = argv[0];

	//01.03.2001 Erwin Vervondel
	//Allow definition of jrepath in ini file
	size = 1 + JSLGetProfileString( _java, "jrepath", "", buffer, bsize-pos, szPath );	
	global_jrePath = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "jrepath", global_jrePath);

	// Create dependencies buffer (a double null terminated array)
	szdependencies = malloc( 1024 );
	memset(szdependencies, 0, 1024);
	size = JSLGetProfileString( _service, "dependencies", "", szdependencies, 1023, szPath );	

	if ( show )
		printf( "%s=%s\n", "dependencies", szdependencies );
	
	for (i = 0; i < size; i++)
	{
		if (szdependencies[i] == ',') 
			szdependencies[i] = 0;

		
	}
	szdependencies[ size ] = 0;

	if ( show ) {
		int pos = 0;
		for (i = 0; i < size+1; i++)
		{
			if (szdependencies[i] == 0) 
			{
				printf( "%s=%s\n", "dependency", szdependencies + pos  );
				pos = i+1;
			}
		}
	}
	//end 01.03.2001 Erwin Vervondel


	dwStartType = SERVICE_AUTO_START;
	//start type
	size = 1 + JSLGetProfileString( _service, "starttype", "auto", buffer, bsize-pos, szPath );
	if ( show )
	  printf( "%s=%s\n", "starttype", buffer);	

	if (strcmp(buffer, "boot") == 0 )
		dwStartType = SERVICE_BOOT_START;

	if (strcmp(buffer, "system") == 0 )
		dwStartType = SERVICE_SYSTEM_START;

	if (strcmp(buffer, "auto") == 0 )
		dwStartType = SERVICE_AUTO_START;

	if (strcmp(buffer, "demand") == 0 )
		dwStartType = SERVICE_DEMAND_START;

	if (strcmp(buffer, "disabled") == 0 )
		dwStartType = SERVICE_DISABLED;

	buffer += size;
	pos += size;
	

	szAccount = NULL;
	//Start account
	size = 1 + JSLGetProfileString( _service, "account", "", buffer, bsize-pos, szPath );
	if ( buffer[0] != 0 )
		szAccount = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "account", szAccount);

	szPwd = NULL;
	//Start password
	size = 1 + JSLGetProfileString( _service, "password", "", buffer, bsize-pos, szPath );
	if ( buffer[0] != 0 )
		szPwd = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "password", szPwd);	

	szLoadOrderGroup = NULL;
	//Load order group
	size = 1 + JSLGetProfileString( _service, "loadordergroup", "", buffer, bsize-pos, szPath );	
	if ( buffer[0] != 0 )
		szLoadOrderGroup = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "loadordergroup", szLoadOrderGroup);

	//Working directory
	size = 1 + JSLGetProfileString( _java, "wrkdir", "", buffer, bsize-pos, szPath );	
	global_wrkdir = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "wrkdir", global_wrkdir);

	//Application name
	size = 1 + JSLGetProfileString( _service, "appname", "appname", buffer, bsize-pos, szPath );	
	szappname = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "appname", szappname );

	//Service name
	size = 1 + JSLGetProfileString( _service, "servicename", "servicename", buffer, bsize-pos, szPath );	
	szservicename = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "servicename", szservicename );

	//Displayname
	size = 1 + JSLGetProfileString( _service, "displayname", "displayname", buffer, bsize-pos, szPath );	
	szdisplayname = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "displayname", szdisplayname );

	//p0.99
	//stdout and stderr redirect
	size = 1 + JSLGetProfileString( _service, "stdout", "", buffer, bsize-pos, szPath );	
	stdoutfile = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "stdout", stdoutfile );

	size = 1 + JSLGetProfileString( _service, "stderr", "", buffer, bsize-pos, szPath );	
	stderrfile = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "stderr", stderrfile );


	//Exception error code
	exceptionErrorCode = GetPrivateProfileInt( _service, "exceptionerrorcode", exceptionErrorCode,  szPath );	
 
	if ( show )
	  printf( "%s=%li\n", "exceptionerrorcode", exceptionErrorCode );

	//Behaviour for non zero exit code
	size = 1 + JSLGetProfileString( _service, "onexiterror", "ignore", buffer, bsize-pos, szPath );	
	onExitError = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "onexiterror", onExitError );


	//Class to use for stopping the service
	size = 1 + JSLGetProfileString( _service, "stopclass", "", buffer, bsize-pos, szPath );	
	stopClass = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "stopclass", stopClass );

	//Method to use for stopping the service
	size = 1 + JSLGetProfileString( _service, "stopmethod", "", buffer, bsize-pos, szPath );	
	stopMethod = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "stopmethod", stopMethod );

	//Signature to use for stopping the service
	size = 1 + JSLGetProfileString( _service, "stopsignature", "", buffer, bsize-pos, szPath );	
	stopSignature = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "stopsignature", stopSignature );


	//Class to use for pausing the service
	size = 1 + JSLGetProfileString( _service, "pauseclass", "", buffer, bsize-pos, szPath );	
	pauseClass = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "pauseclass", pauseClass );

	//Method to use for pausing the service
	size = 1 + JSLGetProfileString( _service, "pausemethod", "", buffer, bsize-pos, szPath );	
	pauseMethod = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "pausemethod", pauseMethod );

	//Signature to use for pausing the service
	size = 1 + JSLGetProfileString( _service, "pausesignature", "", buffer, bsize-pos, szPath );	
	pauseSignature = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "pausesignature", pauseSignature );

	//Class to use for continueing the service
	size = 1 + JSLGetProfileString( _service, "contclass", "", buffer, bsize-pos, szPath );	
	contClass = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "contclass", contClass );

	//Method to use for continueing the service
	size = 1 + JSLGetProfileString( _service, "contmethod", "", buffer, bsize-pos, szPath );	
	contMethod = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "contmethod", contMethod );

	//Signature to use for continueing the service
	size = 1 + JSLGetProfileString( _service, "contsignature", "", buffer, bsize-pos, szPath );	
	contSignature = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "contsignature", contSignature );



	//Stop port
	stopport = GetPrivateProfileInt( _service, "stopport", stopport,  szPath );	
 
	if ( show )
	  printf( "%s=%li\n", "stopport", stopport );

	//Stop port
	startdelay = GetPrivateProfileInt( _service, "startdelay", startdelay,  szPath );	
 
	if ( show )
	  printf( "%s=%li\n", "startdelay", startdelay );
	

	//Using a console handler
	size = 1 + JSLGetProfileString( _service, "useconsolehandler", "yes", buffer, bsize-pos, szPath );	
	szconsolehandler = buffer;
	buffer += size;
	pos += size;

	if ( strcmp( szconsolehandler, "FALSE" ) == 0 )
		useConsoleHandler = FALSE;
	if ( strcmp( szconsolehandler, "false" ) == 0 )
		useConsoleHandler = FALSE;
	if ( strcmp( szconsolehandler, "NO" ) == 0 )
		useConsoleHandler = FALSE;
	if ( strcmp( szconsolehandler, "no" ) == 0 )
		useConsoleHandler = FALSE;

	if ( show )
	  printf( "%s=%s\n", "useconsolehandler", szconsolehandler );

	//Signature to use for stopping the service
	size = 1 + JSLGetProfileString( _service, "path", "", buffer, bsize-pos, szPath );	
	path = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "path", path );

	if ( strlen( path ) != 0 )
		SetEnvironmentVariable( "PATH", path );

	//Executable to run the service from 
	size = 1 + JSLGetProfileString( _service, "modulepath", szModulePath, buffer, bsize-pos, szPath );	
	szModulePath = buffer;
	buffer += size;
	pos += size;
	if ( show )
	  printf( "%s=%s\n", "modulepath", szModulePath );


	//Read first param ...
	size = 1 + JSLGetProfileString( _java, _param, "", buffer, bsize-pos, szPath );	
	stmp = buffer;		
	buffer += size;

	//and check for -classic -hotspot -server -client params
	if (strcmp(stmp, "-classic") == 0 || strcmp(stmp, "-server") == 0 || strcmp(stmp, "-hotspot") == 0 || strcmp(stmp, "-client") == 0 )
	{
	  startParam = 1;
	  _snprintf( commandLine, bsize-cpos, "%s ", stmp );
	  nargv[currentParam++] = stmp;
	  commandLine += size;
	  cpos += size;	  
	  pos += size;
	  if ( show )
		printf( "%s=%s\n", _param, stmp );
	}

	//16.05.01 Erwin Vervondel
	//only set 'service' options if application is run as service
	if (runAsService)
	{		
		//16.7.99 Michael Rschter
		//Made string formatting more robust
		size = 1 + _snprintf( buffer, bsize-pos, "-Dservice.stop.port=%li", stopport );
		_snprintf( commandLine, bsize-cpos, "-Dservice.stop.port=%li ", stopport );
		commandLine += size;
		cpos += size;
		nargv[currentParam++] = buffer;	
		buffer += size;
		pos += size;	

		//New 24.06.1999, Heiner Jostkleigrewe
		//16.7.99 Michael Rschter
		//Made string formatting more robust
		size = 1 + _snprintf( buffer, bsize-pos, "-Dservice.path=%s%s", szDrive, szDir );
		_snprintf( commandLine, bsize-cpos, "-Dservice.path=%s%s ", szDrive, szDir );
		commandLine += size;
		cpos += size;
		nargv[currentParam++] = buffer;
		buffer += size;
		pos += size;

		size = 1 + _snprintf( buffer, bsize-pos, "-Dservice.name=%s", szservicename );
		_snprintf( commandLine, bsize-cpos, "-Dservice.name=%s ", szservicename );
		commandLine += size;
		cpos += size;
		nargv[currentParam++] = buffer;
		buffer += size;
		pos += size;
	
	}	
	//New End
	
	for( i=startParam; i<params; i++)
	{
		_param[5] = 48 + i / 10;
		_param[6] = 48 + i % 10;

		size = 1+JSLGetProfileString( _java, _param, "", buffer, bsize-pos, szPath );	
		_snprintf( commandLine, bsize-cpos, "%s ", buffer );
		commandLine += size;
		cpos += size;

		nargv[ currentParam++ ] = buffer;
		buffer += size;
		pos += size;

		if ( show )
		  printf( "%s=%s\n", _param, nargv[ currentParam - 1  ] );
	}
	if ( show )
	  printf( "Java command line:\njava %s\n", cLine );


	if ( bDebug )
		logf( cLine );
}

void getIniFile( LPTSTR szPath )
{
	int i;
    if ( GetModuleFileName( NULL, szPath, 512 ) == 0)
		AddToMessageLog(TEXT("Could not get module name, so in consequence could not locate ini file."));	
	//construct the ini file name 	
	
	for( i=0; szPath[i] != 0; i++ );
	szPath[i-3] = 'i';
	szPath[i-2] = 'n';
	szPath[i-1] = 'i';
}



int JSLGetProfileString( LPTSTR section, LPTSTR name, LPTSTR _default, LPTSTR buffer, int bsize, LPTSTR path )
{
	DWORD size = bsize;
	LPTSTR tmpbuf = NULL;
	LPTSTR start;
	LPTSTR stop;
	LPTSTR buf;

	GetPrivateProfileString( section, name, _default, buffer, bsize, path );
	buf = buffer;
	start = strchr( buf,'%');

	//printf( "Parsing %s\n", buf );

	//no buffer if there is no substitution
	if ( start  )
	{
		//printf( "found %% \n");
		tmpbuf = malloc( bsize );
		tmpbuf[0] = 0;
	}

	while ( start)
	{
		//printf( "A %s\n", buf );
		//printf( "B %s\n", tmpbuf );

		stop = strchr( start+1, '%' );
		if ( stop )
		{
			start[0] = 0;
			stop[0] = 0;
			
			//printf( "T %s\n", start+1 );
			if ( size >= strlen( buf ) )
			{
				strcat( tmpbuf, buf );
				size -= strlen(buf);
			}
			//Convert %% in to single %
			if ( stop - start == 1 )
			{
				if ( size >= 1 )
				{
					strcat( tmpbuf, "%" );
					size--;
				}
			} else
			{
				size -= GetEnvironmentVariable( start+1, tmpbuf + strlen(tmpbuf) , size );				
			}
			buf = stop + 1;
			start = strchr( buf,'%');			

		} else
		{
			start = NULL;
		}
	}	

	if ( tmpbuf)
	{
		//printf( "C %s\n", buf );
		//printf( "D %s\n", tmpbuf );

		strcat( tmpbuf, buf );
		strcpy( buffer, tmpbuf );
		free( tmpbuf );
	}

	//printf( "E %s\n", buffer );

	//printf( "End parsing\n" );
	return strlen( buffer );
}


void getLogFile( LPTSTR szPath )
{
	int i;
    if ( GetModuleFileName( NULL, szPath, 512 ) == 0)
		AddToMessageLog(TEXT("Could not get module name, so in consequence could not locate log file."));	
	//construct the ini file name 	
	
	for( i=0; szPath[i] != 0; i++ );
	szPath[i-3] = 'l';
	szPath[i-2] = 'o';
	szPath[i-1] = 'g';
}

/**
  * Log to file
*/
void logf( LPTSTR s )
{	
	TCHAR szPath[512];
	FILE* f;
	long theTime;

	theTime = time( NULL );

	getLogFile( szPath );
	f = fopen( szPath, "a+" );
	
	fprintf( f, ":%d:%s\n" , theTime, s );

	fclose( f );
}

void logfd( long d )
{	
	TCHAR szPath[512];
	FILE* f;
	long theTime;

	theTime = time( NULL );

	getLogFile( szPath );
	f = fopen( szPath, "a+" );
	
	fprintf( f, ":%d:%d\n" , theTime, d );

	fclose( f );
}



// New 24.06.1999, Heiner Jostkleigrewe
void getIniDir( LPTSTR szPath , LPTSTR szDrive, LPTSTR szDir)
{
   TCHAR fname[_MAX_FNAME];
   TCHAR ext[_MAX_EXT];

   _splitpath( szPath, szDrive, szDir, fname, ext );
}
// New Ende

/**
  * Perform service initialization
 */
void init()
{
	// register a service control handler
    sshStatusHandle = RegisterServiceCtrlHandler( TEXT(SZSERVICENAME), service_ctrl);

}

//
//  FUNCTION: service_main
//
//  PURPOSE: To perform actual initialization of the service
//
//  PARAMETERS:
//    dwArgc   - number of command line arguments
//    lpszArgv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    This routine performs the service initialization and then calls
//    the user defined ServiceStart() routine to perform majority
//    of the work.
//
void WINAPI service_main(DWORD dwArgc, LPTSTR *lpszArgv)
{
	if ( bDebug )
		logf( "service main called" );
	
    init();

    if (!sshStatusHandle)
        goto cleanup;

    // Some 
    ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    ssStatus.dwServiceSpecificExitCode = 0;

	if ( bDebug )
		logf( "OK 1" );

    // report the status to the service control manager.
    if (!ReportStatusToSCMgr(
		SERVICE_START_PENDING, // service state
        NO_ERROR,              // exit code
        60000))                 // wait hint
        goto cleanup;
	
	
	//p0.99
	prepare();
    ServiceStart( global_argc, global_argv );

cleanup:

	if ( bDebug )
		logf( "service cleanup" );
	

    // try to report the stopped status to the service control manager.
    if (sshStatusHandle)
        (VOID)ReportStatusToSCMgr(
                            SERVICE_STOPPED,
                            dwErr,
                            0);

	if ( bDebug )
		logf( "service exiting" );
	

    return;
}


//
//  FUNCTION: service_ctrl
//
//  PURPOSE: This function is called by the SCM whenever
//           ControlService() is called on this service.
//
//  PARAMETERS:
//    dwCtrlCode - type of control requested
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
VOID WINAPI service_ctrl(DWORD dwCtrlCode)
{
	if ( bDebug )
		logf( "service control call" );
    // Handle the requested control code.
    //
    switch(dwCtrlCode)
    {
        // Stop the service.
        //
        // SERVICE_STOP_PENDING should be reported before
        // setting the Stop Event - hServerStopEvent - in
        // ServiceStop().  This avoids a race condition
        // which may result in a 1053 - The Service did not respond...
        // error.
        case SERVICE_CONTROL_STOP:
		case SERVICE_CONTROL_SHUTDOWN:
            ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 60000);			
            ServiceStop();	
			return;
			
		case SERVICE_CONTROL_PAUSE: 
				
			ServicePause();
            return; 

		case SERVICE_CONTROL_CONTINUE: 			 
			ReportStatusToSCMgr(SERVICE_CONTINUE_PENDING, NO_ERROR, 60000);
			ServiceContinue();
            return; 

        // Update the service status.
        //
        case SERVICE_CONTROL_INTERROGATE:
            break;

        // invalid control code
        //
        default:
            break;

    }

    ReportStatusToSCMgr(ssStatus.dwCurrentState, NO_ERROR, 0);
}



//
//  FUNCTION: ReportStatusToSCMgr()
//
//  PURPOSE: Sets the current status of the service and
//           reports it to the Service Control Manager
//
//  PARAMETERS:
//    dwCurrentState - the state of the service
//    dwWin32ExitCode - error code to report
//    dwWaitHint - worst case estimate to next checkpoint
//
//  RETURN VALUE:
//    TRUE  - success
//    FALSE - failure
//
//  COMMENTS:
//
BOOL ReportStatusToSCMgr(DWORD dwCurrentState,
                         DWORD dwWin32ExitCode,
                         DWORD dwWaitHint)
{
    static DWORD dwCheckPoint = 1;
    BOOL fResult = TRUE;

    if ( !bCommandLine ) // when debugging we don't report to the SCM
    {
        if (dwCurrentState == SERVICE_START_PENDING)
            ssStatus.dwControlsAccepted = 0;
        else
		{
            ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;
			if ( strlen( pauseClass ) && strlen( pauseMethod ) && strlen( pauseSignature ) && 
				 strlen( contClass ) && strlen( contMethod ) && strlen( contSignature ) )
			{
				ssStatus.dwControlsAccepted |= SERVICE_ACCEPT_PAUSE_CONTINUE;
			}
		}
        ssStatus.dwCurrentState = dwCurrentState;

		//p0.99
		if ( dwCurrentState == SERVICE_STOPPED && dwWin32ExitCode != 0 && strcmp( onExitError, "reportcode" )==0 )
		{
			ssStatus.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
			ssStatus.dwServiceSpecificExitCode = dwWin32ExitCode;
        } else
			ssStatus.dwWin32ExitCode = 0;

		
		if ( dwCurrentState == SERVICE_STOPPED && strcmp( onExitError, "fatal" )==0 && dwWin32ExitCode != 0 )
		{
			//printf( "Fatal exit code %d", dwWin32ExitCode );
			exit( dwWin32ExitCode );
		}


        ssStatus.dwWaitHint = dwWaitHint;

        if ( ( dwCurrentState == SERVICE_RUNNING ) ||
             ( dwCurrentState == SERVICE_STOPPED ) )
            ssStatus.dwCheckPoint = 0;
        else
            ssStatus.dwCheckPoint = dwCheckPoint++;


        // Report the status of the service to the service control manager.
        //
        if (!(fResult = SetServiceStatus( sshStatusHandle, &ssStatus))) {
            AddToMessageLog(TEXT("SetServiceStatus"));
        }
    }
    return fResult;
}

//
//  FUNCTION: AddToMessageLog(LPTSTR lpszMsg)
//
//  PURPOSE: Allows any thread to log an error message
//
//  PARAMETERS:
//    lpszMsg - text for message
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
VOID AddToMessageLog(LPTSTR lpszMsg)
{
    TCHAR   szMsg[256];
    HANDLE  hEventSource;
    LPTSTR  lpszStrings[2];


    if ( !bDebug )
    {
        dwErr = GetLastError();

        // Use event logging to log the error.
        //
        hEventSource = RegisterEventSource(NULL, TEXT(SZSERVICENAME));

        _stprintf(szMsg, TEXT("%s error: %d"), TEXT(SZSERVICENAME), dwErr);
        lpszStrings[0] = szMsg;
        lpszStrings[1] = lpszMsg;

        if (hEventSource != NULL) {
            ReportEvent(hEventSource, // handle of event source
                EVENTLOG_ERROR_TYPE,  // event type
                0,                    // event category
                0,                    // event ID
                NULL,                 // current user's SID
                2,                    // strings in lpszStrings
                0,                    // no bytes of raw data
                lpszStrings,          // array of error strings
                NULL);                // no raw data

            (VOID) DeregisterEventSource(hEventSource);
        }
    }
}




///////////////////////////////////////////////////////////////////
//
//  The following code handles service installation and removal
//


//
//  FUNCTION: CmdConfigureService()
//
//  PURPOSE: Installs the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//

void CmdConfigureService()
{
    SC_HANDLE   schService;
    SC_HANDLE   schSCManager;
	BOOL ret;

    TCHAR szPath[512];

    //p0.99
	if ( strlen(szModulePath) == 0 )
	{
		if ( GetModuleFileName( NULL, szPath, 512 ) == 0 )
		{
			_tprintf(TEXT("Unable to configure %s - %s\n"), TEXT(SZSERVICEDISPLAYNAME), GetLastErrorText(szErr, 256));
			return;
		}
		szModulePath = szPath;
	}

    schSCManager = OpenSCManager(
                        NULL,                   // machine (NULL == local)
                        NULL,                   // database (NULL == default)
                        SC_MANAGER_ALL_ACCESS   // access required
                        );
    if ( schSCManager )
    {
		schService = OpenService( 
			schSCManager,				// handle to service control manager database 
			TEXT(SZSERVICENAME),		// pointer to name of service to start 
			SERVICE_ALL_ACCESS			// type of access to service
		); 

		if ( schService )
        {
        }
        else
        {
            _tprintf(TEXT("OpenService failed - %s\n"), GetLastErrorText(szErr, 256));
			CloseServiceHandle(schSCManager);
			return;
        }


        ret =  ChangeServiceConfig(
            schService,					// SCManager database
            SERVICE_WIN32_OWN_PROCESS,  // service type
            dwStartType,				// start type
            SERVICE_ERROR_NORMAL,       // error control type
            szModulePath,                     // service's binary
            szLoadOrderGroup,           // load ordering group
            NULL,                       // no tag identifier
            szdependencies,				// dependencies
            szAccount,                  // account
            szPwd,						// password
			TEXT(SZSERVICEDISPLAYNAME)
			);                      

        if ( ret )
        {
            _tprintf(TEXT("%s reconfigured.\n"), TEXT(SZSERVICEDISPLAYNAME) );
            CloseServiceHandle(schService);
        }
        else
        {
            _tprintf(TEXT("CreateService failed - %s\n"), GetLastErrorText(szErr, 256));
        }

        CloseServiceHandle(schSCManager);
    }
    else
        _tprintf(TEXT("OpenSCManager failed - %s\n"), GetLastErrorText(szErr,256));
}

//
void CmdInstallService()
{
    SC_HANDLE   schService;
    SC_HANDLE   schSCManager;

    TCHAR szPath[512];

	//p0.99
	if ( strlen(szModulePath) == 0 )
	{
		if ( GetModuleFileName( NULL, szPath, 512 ) == 0 )
		{
			_tprintf(TEXT("Unable to configure %s - %s\n"), TEXT(SZSERVICEDISPLAYNAME), GetLastErrorText(szErr, 256));
			return;
		}
		szModulePath = szPath;
	}
		

    schSCManager = OpenSCManager(
                        NULL,                   // machine (NULL == local)
                        NULL,                   // database (NULL == default)
                        SC_MANAGER_ALL_ACCESS   // access required
                        );
    if ( schSCManager )
    {
        schService = CreateService(
            schSCManager,               // SCManager database
            TEXT(SZSERVICENAME),        // name of service
            TEXT(SZSERVICEDISPLAYNAME), // name to display
            SERVICE_ALL_ACCESS | GENERIC_EXECUTE  ,         // desired access
            SERVICE_WIN32_OWN_PROCESS,  // service type
            dwStartType,				// start type
            SERVICE_ERROR_NORMAL,       // error control type
			//p0.99
            szModulePath,                     // service's binary
            szLoadOrderGroup,			// load ordering group
            NULL,                       // no tag identifier
            szdependencies,				// dependencies
            szAccount,                  // account
            szPwd);                     // password

        if ( schService )
        {
            _tprintf(TEXT("%s installed as a Windows service.\n"), TEXT(SZSERVICEDISPLAYNAME) );
            CloseServiceHandle(schService);
        }
        else
        {
            _tprintf(TEXT("CreateService failed - %s\n"), GetLastErrorText(szErr, 256));
        }

        CloseServiceHandle(schSCManager);
    }
    else
        _tprintf(TEXT("OpenSCManager failed - %s\n"), GetLastErrorText(szErr,256));
}


//
//  FUNCTION: CmdRemoveService()
//
//  PURPOSE: Stops and removes the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
void CmdRemoveService()
{
    SC_HANDLE   schService;
    SC_HANDLE   schSCManager;

    schSCManager = OpenSCManager(
                        NULL,                   // machine (NULL == local)
                        NULL,                   // database (NULL == default)
                        SC_MANAGER_ALL_ACCESS   // access required
                        );
    if ( schSCManager )
    {
        schService = OpenService(schSCManager, TEXT(SZSERVICENAME), SERVICE_ALL_ACCESS);

        if (schService)
        {
            // try to stop the service
            if ( ControlService( schService, SERVICE_CONTROL_STOP, &ssStatus ) )
            {
                _tprintf(TEXT("Stopping %s."), TEXT(SZSERVICEDISPLAYNAME));
                Sleep( 1000 );

                while( QueryServiceStatus( schService, &ssStatus ) )
                {
                    if ( ssStatus.dwCurrentState == SERVICE_STOP_PENDING )
                    {
                        _tprintf(TEXT("."));
                        Sleep( 1000 );
                    }
                    else
                        break;
                }

                if ( ssStatus.dwCurrentState == SERVICE_STOPPED )
                    _tprintf(TEXT("\n%s stopped.\n"), TEXT(SZSERVICEDISPLAYNAME) );
                else
                    _tprintf(TEXT("\n%s failed to stop.\n"), TEXT(SZSERVICEDISPLAYNAME) );

            }

            // now remove the service
            if( DeleteService(schService) )
                _tprintf(TEXT("%s removed as a Windows service.\n"), TEXT(SZSERVICEDISPLAYNAME) );
            else
                _tprintf(TEXT("DeleteService failed - %s\n"), GetLastErrorText(szErr,256));


            CloseServiceHandle(schService);
        }
        else
            _tprintf(TEXT("OpenService failed - %s\n"), GetLastErrorText(szErr,256));

        CloseServiceHandle(schSCManager);
    }
    else
        _tprintf(TEXT("OpenSCManager failed - %s\n"), GetLastErrorText(szErr,256));
}




///////////////////////////////////////////////////////////////////
//
//  The following code is for running the service as a console app
//


//
//  FUNCTION: CmdDebugService(int argc, char ** argv)
//
//  PURPOSE: Runs the service as a console application
//
//  PARAMETERS:
//    argc - number of command line arguments
//    argv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
void CmdDebugService(int argc, LPTSTR* argv)
{
    DWORD dwArgc;
    LPTSTR *lpszArgv;

#ifdef UNICODE
    lpszArgv = CommandLineToArgvW(GetCommandLineW(), &(dwArgc) );
#else
    dwArgc   = (DWORD) argc;
    lpszArgv = argv;
#endif

    _tprintf(TEXT("Debugging %s.\n"), TEXT(SZSERVICEDISPLAYNAME));

    SetConsoleCtrlHandler( ControlHandler, TRUE );

	//p0.99
	prepare();
    ServiceStart( dwArgc, lpszArgv );
}

///////////////////////////////////////////////////////////////////
//
//  The following code is for running the service as a console app
//

//
//  FUNCTION: CmdRunService(int argc, char ** argv)
//
//  PURPOSE: Runs the service as a console application
//
//  PARAMETERS:
//    argc - number of command line arguments
//    argv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
void CmdRunService(int argc, LPTSTR* argv)
{
    DWORD dwArgc;
    LPTSTR *lpszArgv;

#ifdef UNICODE
    lpszArgv = CommandLineToArgvW(GetCommandLineW(), &(dwArgc) );
#else
    dwArgc   = (DWORD) argc;
    lpszArgv = argv;
#endif

    _tprintf(TEXT("Running %s.\n"), TEXT(SZSERVICEDISPLAYNAME));

    SetConsoleCtrlHandler( ControlHandler, TRUE );

	//p0.99
	prepare();
    ServiceStart( dwArgc, lpszArgv );
}


//
//  FUNCTION: ControlHandler ( DWORD dwCtrlType )
//
//  PURPOSE: Handled console control events
//
//  PARAMETERS:
//    dwCtrlType - type of control event
//
//  RETURN VALUE:
//    True - handled
//    False - unhandled
//
//  COMMENTS:
//
BOOL WINAPI ControlHandler ( DWORD dwCtrlType )
{
    switch( dwCtrlType )
    {
        case CTRL_BREAK_EVENT:  // use Ctrl+Break to simulate service pause and continue
			
			if ( bPaused )
			{	
				ServiceContinue();
				bPaused = FALSE;
			} else
			{
				ServicePause();
				bPaused = TRUE;
			}
			
			return TRUE;
			break;

        case CTRL_C_EVENT:   // SERVICE_CONTROL_STOP in debug mode
            _tprintf(TEXT("Stopping %s.\n"), TEXT(SZSERVICEDISPLAYNAME));
            
			ServiceStop();
            return TRUE;
            break;

     
    }
    return FALSE;
}

void prepare()
{
	//p0.99
	//redirect stdout and stderr
	if ( strlen( stdoutfile ) != 0 )
		restdout = freopen( stdoutfile, "a", stdout );
	if ( strlen( stderrfile ) != 0 )
		restderr = freopen( stderrfile, "a", stderr );
}

void cleanup()
{
	if ( restdout )
	{
		fclose( restdout );
		restdout = NULL;
	}
	if ( restderr )
	{
		fclose( restderr );
		restderr = NULL;
	}
}

//
//  FUNCTION: GetLastErrorText
//
//  PURPOSE: copies error message text to string
//
//  PARAMETERS:
//    lpszBuf - destination buffer
//    dwSize - size of buffer
//
//  RETURN VALUE:
//    destination buffer
//
//  COMMENTS:
//
LPTSTR GetLastErrorText( LPTSTR lpszBuf, DWORD dwSize )
{
    DWORD dwRet;
    LPTSTR lpszTemp = NULL;

    dwRet = FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |FORMAT_MESSAGE_ARGUMENT_ARRAY,
                           NULL,
                           GetLastError(),
                           LANG_NEUTRAL,
                           (LPTSTR)&lpszTemp,
                           0,
                           NULL );

    // supplied buffer is not long enough
    if ( !dwRet || ( (long)dwSize < (long)dwRet+14 ) )
        lpszBuf[0] = TEXT('\0');
    else
    {
        lpszTemp[lstrlen(lpszTemp)-2] = TEXT('\0');  //remove cr and newline character
        _stprintf( lpszBuf, TEXT("%s (0x%x)"), lpszTemp, GetLastError() );
    }

    if ( lpszTemp )
        LocalFree((HLOCAL) lpszTemp );

    return lpszBuf;
}
